@@ -1,7 +1,7 @@
 /*
  * The XML and API file generator module for SIP.
  *
- * Copyright (c) 2019 Riverbank Computing Limited <info@riverbankcomputing.com>
+ * Copyright (c) 2010 Riverbank Computing Limited <info@riverbankcomputing.com>
  *
  * This file is part of SIP.
  *
@@ -34,41 +34,31 @@
 
 static void apiEnums(sipSpec *pt, moduleDef *mod, classDef *scope, FILE *fp);
 static void apiVars(sipSpec *pt, moduleDef *mod, classDef *scope, FILE *fp);
-static void apiCtor(sipSpec *pt, moduleDef *mod, classDef *scope, ctorDef *ct,
-        FILE *fp);
-static void apiOverload(sipSpec *pt, moduleDef *mod, classDef *scope,
-        overDef *od, FILE *fp);
+static int apiCtor(sipSpec *pt, moduleDef *mod, classDef *scope, ctorDef *ct,
+        int sec, FILE *fp);
+static int apiOverload(sipSpec *pt, moduleDef *mod, classDef *scope,
+        overDef *od, int sec, FILE *fp);
 static int apiArgument(sipSpec *pt, argDef *ad, int out, int need_comma,
-        int names, int defaults, int in_str, FILE *fp);
+        int sec, int names, int defaults, int in_str, FILE *fp);
 static void xmlClass(sipSpec *pt, moduleDef *mod, classDef *cd, FILE *fp);
 static void xmlEnums(sipSpec *pt, moduleDef *mod, classDef *scope, int indent,
         FILE *fp);
 static void xmlVars(sipSpec *pt, moduleDef *mod, classDef *scope, int indent,
         FILE *fp);
-static void xmlFunction(sipSpec *pt, moduleDef *mod, classDef *scope,
-        memberDef *md, overDef *oloads, int indent, FILE *fp);
-static void xmlCtor(sipSpec *pt, moduleDef *mod, classDef *scope, ctorDef *ct,
+static void xmlFunction(sipSpec *pt, classDef *scope, memberDef *md,
+        overDef *oloads, int indent, FILE *fp);
+static int xmlCtor(sipSpec *pt, classDef *scope, ctorDef *ct, int sec,
         int indent, FILE *fp);
-static void xmlOverload(sipSpec *pt, moduleDef *mod, classDef *scope,
-        memberDef *md, overDef *od, classDef *xtnds, int stat, int indent,
-        FILE *fp);
-static void xmlCppSignature(FILE *fp, signatureDef *sd, int is_const);
-static void xmlArgument(sipSpec *pt, moduleDef *mod, argDef *ad, int out,
-        KwArgs kwargs, int res_xfer, int indent, FILE *fp);
-static void xmlType(sipSpec *pt, moduleDef *mod, argDef *ad, int out,
-        KwArgs kwargs, FILE *fp);
+static int xmlOverload(sipSpec *pt, classDef *scope, memberDef *md,
+        overDef *od, classDef *xtnds, int stat, int sec, int indent, FILE *fp);
+static void xmlCppSignature(FILE *fp, overDef *od);
+static void xmlArgument(sipSpec *pt, argDef *ad, const char *dir, int res_xfer,
+        int sec, int indent, FILE *fp);
+static void xmlType(sipSpec *pt, argDef *ad, int sec, FILE *fp);
 static void xmlIndent(int indent, FILE *fp);
-static void xmlRealName(scopedNameDef *fqcname, const char *member, FILE *fp);
-static void xmlRealScopedName(classDef *scope, const char *cppname, FILE *fp);
-static const char *pyType(sipSpec *pt, argDef *ad, classDef **scope);
-static void exportPythonSignature(sipSpec *pt, FILE *fp, signatureDef *sd,
-        int names, int defaults, int in_str, int is_signal);
-static void restPyEnumMember(enumMemberDef *emd, FILE *fp);
-static void restPyAttribute(moduleDef *mod, classDef *scope, nameDef *name,
-        FILE *fp);
-static int restValue(sipSpec *pt, valueDef *value, FILE *fp);
-static const char *reflectedSlot(slotType st);
-static int hasCppSignature(signatureDef *sd);
+static const char *dirAttribute(argDef *ad);
+static void exportDefaultValue(argDef *ad, int in_str, FILE *fp);
+static const char *pyType(sipSpec *pt, argDef *ad, int sec, classDef **scope);
 
 
 /*
@@ -95,7 +85,8 @@ void generateAPI(sipSpec *pt, moduleDef *mod, const char *apiFile)
         if (od->common->slot != no_slot)
             continue;
 
-        apiOverload(pt, mod, NULL, od, fp);
+        if (apiOverload(pt, mod, NULL, od, FALSE, fp))
+            apiOverload(pt, mod, NULL, od, TRUE, fp);
     }
 
     for (cd = pt->classes; cd != NULL; cd = cd->next)
@@ -116,7 +107,8 @@ void generateAPI(sipSpec *pt, moduleDef *mod, const char *apiFile)
             if (isPrivateCtor(ct))
                 continue;
 
-            apiCtor(pt, mod, cd, ct, fp);
+            if (apiCtor(pt, mod, cd, ct, FALSE, fp))
+                apiCtor(pt, mod, cd, ct, TRUE, fp);
         }
 
         for (od = cd->overs; od != NULL; od = od->next)
@@ -127,7 +119,8 @@ void generateAPI(sipSpec *pt, moduleDef *mod, const char *apiFile)
             if (od->common->slot != no_slot)
                 continue;
 
-            apiOverload(pt, mod, cd, od, fp);
+            if (apiOverload(pt, mod, cd, od, FALSE, fp))
+                apiOverload(pt, mod, cd, od, TRUE, fp);
         }
     }
 
@@ -138,10 +131,10 @@ void generateAPI(sipSpec *pt, moduleDef *mod, const char *apiFile)
 /*
  * Generate an API ctor.
  */
-static void apiCtor(sipSpec *pt, moduleDef *mod, classDef *scope, ctorDef *ct,
-        FILE *fp)
+static int apiCtor(sipSpec *pt, moduleDef *mod, classDef *scope, ctorDef *ct,
+        int sec, FILE *fp)
 {
-    int need_comma, a;
+    int need_sec = FALSE, need_comma, a;
 
     /* Do the callable type form. */
     fprintf(fp, "%s.", mod->name);
@@ -154,8 +147,11 @@ static void apiCtor(sipSpec *pt, moduleDef *mod, classDef *scope, ctorDef *ct,
     {
         argDef *ad = &ct->pysig.args[a];
 
-        need_comma = apiArgument(pt, ad, FALSE, need_comma, TRUE, TRUE, FALSE,
-                fp);
+        need_comma = apiArgument(pt, ad, FALSE, need_comma, sec, TRUE, TRUE,
+                FALSE, fp);
+
+        if (ad->atype == rxcon_type || ad->atype == rxdis_type)
+            need_sec = TRUE;
     }
 
     fprintf(fp, ")\n");
@@ -166,10 +162,12 @@ static void apiCtor(sipSpec *pt, moduleDef *mod, classDef *scope, ctorDef *ct,
     fprintf(fp, ".__init__?%d(self", CLASS_ID);
 
     for (a = 0; a < ct->pysig.nrArgs; ++a)
-        apiArgument(pt, &ct->pysig.args[a], FALSE, TRUE, TRUE, TRUE, FALSE,
-                fp);
+        apiArgument(pt, &ct->pysig.args[a], FALSE, TRUE, sec, TRUE, TRUE,
+                FALSE, fp);
 
     fprintf(fp, ")\n");
+
+    return need_sec;
 }
 
 
@@ -232,16 +230,20 @@ static void apiVars(sipSpec *pt, moduleDef *mod, classDef *scope, FILE *fp)
 /*
  * Generate a single API overload.
  */
-static void apiOverload(sipSpec *pt, moduleDef *mod, classDef *scope,
-        overDef *od, FILE *fp)
+static int apiOverload(sipSpec *pt, moduleDef *mod, classDef *scope,
+        overDef *od, int sec, FILE *fp)
 {
+    int need_sec;
+
     fprintf(fp, "%s.", mod->name);
     prScopedPythonName(fp, scope, od->common->pyname->text);
     fprintf(fp, "?%d", METHOD_ID);
 
-    exportPythonSignature(pt, fp, &od->pysig, TRUE, TRUE, FALSE, FALSE);
+    need_sec = prPythonSignature(pt, fp, &od->pysig, sec, TRUE, TRUE, FALSE);
 
     fprintf(fp, "\n");
+
+    return need_sec;
 }
 
 
@@ -249,7 +251,7 @@ static void apiOverload(sipSpec *pt, moduleDef *mod, classDef *scope,
  * Generate the API for an argument.
  */
 static int apiArgument(sipSpec *pt, argDef *ad, int out, int need_comma,
-        int names, int defaults, int in_str, FILE *fp)
+        int sec, int names, int defaults, int in_str, FILE *fp)
 {
     const char *tname;
     classDef *tscope;
@@ -257,7 +259,10 @@ static int apiArgument(sipSpec *pt, argDef *ad, int out, int need_comma,
     if (isArraySize(ad))
         return need_comma;
 
-    if ((tname = pyType(pt, ad, &tscope)) == NULL)
+    if (sec && (ad->atype == slotcon_type || ad->atype == slotdis_type))
+        return need_comma;
+
+    if ((tname = pyType(pt, ad, sec, &tscope)) == NULL)
         return need_comma;
 
     if (need_comma)
@@ -275,7 +280,9 @@ static int apiArgument(sipSpec *pt, argDef *ad, int out, int need_comma,
             fprintf(fp, " %s", ad->name->text);
 
         fprintf(fp, "=");
-        prDefaultValue(ad, in_str, fp);
+        prcode(fp, "%M");
+        exportDefaultValue(ad, in_str, fp);
+        prcode(fp, "%M");
     }
 
     return TRUE;
@@ -320,7 +327,7 @@ void generateXML(sipSpec *pt, moduleDef *mod, const char *xmlFile)
     xmlVars(pt, mod, NULL, 1, fp);
 
     for (md = mod->othfuncs; md != NULL; md = md->next)
-        xmlFunction(pt, mod, NULL, md, mod->overs, 1, fp);
+        xmlFunction(pt, NULL, md, mod->overs, 1, fp);
 
     fprintf(fp, "</Module>\n");
 
@@ -328,54 +335,6 @@ void generateXML(sipSpec *pt, moduleDef *mod, const char *xmlFile)
 }
 
 
-/*
- * Generate a 'realname' attribute containing a fully qualified C/C++ name of
- * an object.
- */
-static void xmlRealScopedName(classDef *scope, const char *cppname, FILE *fp)
-{
-    const char *sep = "";
-
-    fprintf(fp, " realname=\"");
-
-    if (scope != NULL)
-    {
-        scopedNameDef *snd;
-
-        for (snd = removeGlobalScope(classFQCName(scope)); snd != NULL; snd = snd->next)
-        {
-            fprintf(fp, "%s%s", sep, snd->name);
-            sep = "::";
-        }
-    }
-
-    fprintf(fp, "%s%s\"", sep, cppname);
-}
-
-
-/*
- * Generate a 'realname' attribute containing a fully qualified C/C++ name.
- */
-static void xmlRealName(scopedNameDef *fqcname, const char *member, FILE *fp)
-{
-    const char *sep = "";
-    scopedNameDef *snd;
-
-    fprintf(fp, " realname=\"");
-
-    for (snd = removeGlobalScope(fqcname); snd != NULL; snd = snd->next)
-    {
-        fprintf(fp, "%s%s", sep, snd->name);
-        sep = "::";
-    }
-
-    if (member != NULL)
-        fprintf(fp, "::%s", member);
-
-    fprintf(fp, "\"");
-}
-
-
 /*
  * Generate the XML for a class.
  */
@@ -395,83 +354,56 @@ static void xmlClass(sipSpec *pt, moduleDef *mod, classDef *cd, FILE *fp)
         return;
     }
 
-    if (!isHiddenNamespace(cd))
+    xmlIndent(indent++, fp);
+    fprintf(fp, "<Class name=\"");
+    prScopedPythonName(fp, cd->ecd, cd->pyname->text);
+    fprintf(fp, "\"");
+
+    if (cd->picklecode != NULL)
+        fprintf(fp, " pickle=\"1\"");
+
+    if (cd->convtocode != NULL)
+        fprintf(fp, " convert=\"1\"");
+
+    if (cd->real != NULL)
+        fprintf(fp, " extends=\"%s\"", cd->real->iff->module->name);
+
+    if (cd->supers != NULL)
     {
-        xmlIndent(indent++, fp);
-        fprintf(fp, "<Class name=\"");
-        prScopedPythonName(fp, cd->ecd, cd->pyname->text);
+        classList *cl;
+
+        fprintf(fp, " inherits=\"");
+
+        for (cl = cd->supers; cl != NULL; cl = cl->next)
+        {
+            if (cl != cd->supers)
+                fprintf(fp, " ");
+
+            prScopedPythonName(fp, cl->cd->ecd, cl->cd->pyname->text);
+        }
+
         fprintf(fp, "\"");
-
-        xmlRealName(classFQCName(cd), NULL, fp);
-
-        if (cd->picklecode != NULL)
-            fprintf(fp, " pickle=\"1\"");
-
-        if (cd->convtocode != NULL)
-            fprintf(fp, " convert=\"1\"");
-
-        if (cd->convfromcode != NULL)
-            fprintf(fp, " convertfrom=\"1\"");
-
-        if (cd->real != NULL)
-            fprintf(fp, " extends=\"%s\"", cd->real->iff->module->name);
-
-        if (cd->pyqt_flags_enums != NULL)
-        {
-            const char *sep;
-            stringList *sl;
-
-            fprintf(fp, " flagsenums=\"");
-            sep = "";
-
-            for (sl = cd->pyqt_flags_enums; sl != NULL; sl = sl->next)
-            {
-                fprintf(fp, "%s%s", sep, sl->s);
-                sep = " ";
-            }
-
-            fprintf(fp, "\"");
-        }
-
-        if (cd->supers != NULL)
-        {
-            classList *cl;
-
-            fprintf(fp, " inherits=\"");
-
-            for (cl = cd->supers; cl != NULL; cl = cl->next)
-            {
-                if (cl != cd->supers)
-                    fprintf(fp, " ");
-
-                restPyClass(cl->cd, fp);
-            }
-
-            fprintf(fp, "\"");
-        }
-
-        fprintf(fp, ">\n");
     }
 
+    fprintf(fp, ">\n");
+
+    xmlEnums(pt, mod, cd, indent, fp);
+    xmlVars(pt, mod, cd, indent, fp);
+
     for (ct = cd->ctors; ct != NULL; ct = ct->next)
     {
         if (isPrivateCtor(ct))
             continue;
 
-        xmlCtor(pt, mod, cd, ct, indent, fp);
+        if (xmlCtor(pt, cd, ct, FALSE, indent, fp))
+            xmlCtor(pt, cd, ct, TRUE, indent, fp);
     }
 
-    xmlEnums(pt, mod, cd, indent, fp);
-    xmlVars(pt, mod, cd, indent, fp);
-
     for (md = cd->members; md != NULL; md = md->next)
-        xmlFunction(pt, mod, cd, md, cd->overs, indent, fp);
+        xmlFunction(pt, cd, md, cd->overs, indent, fp);
 
-    if (!isHiddenNamespace(cd))
-    {
-        xmlIndent(--indent, fp);
-        fprintf(fp, "</Class>\n");
-    }
+    xmlIndent(--indent, fp);
+    fprintf(fp, "</Class>\n");
 }
 
 
@@ -498,22 +430,14 @@ static void xmlEnums(sipSpec *pt, moduleDef *mod, classDef *scope, int indent,
             xmlIndent(indent++, fp);
             fprintf(fp, "<Enum name=\"");
             prScopedPythonName(fp, ed->ecd, ed->pyname->text);
-            fprintf(fp, "\"");
-
-            xmlRealName(ed->fqcname, NULL, fp);
-
-            fprintf(fp, ">\n");
+            fprintf(fp, "\">\n");
 
             for (emd = ed->members; emd != NULL; emd = emd->next)
             {
                 xmlIndent(indent, fp);
                 fprintf(fp, "<EnumMember name=\"");
-                prScopedPythonName(fp, ed->ecd, ed->pyname->text);
-                fprintf(fp, ".%s\"", emd->pyname->text);
-
-                xmlRealName(ed->fqcname, emd->cname, fp);
-
-                fprintf(fp, "/>\n");
+                prScopedPythonName(fp, ed->ecd, emd->pyname->text);
+                fprintf(fp, "\"/>\n");
             }
 
             xmlIndent(--indent, fp);
@@ -528,11 +452,7 @@ static void xmlEnums(sipSpec *pt, moduleDef *mod, classDef *scope, int indent,
                 xmlIndent(indent, fp);
                 fprintf(fp, "<Member name=\"");
                 prScopedPythonName(fp, ed->ecd, emd->pyname->text);
-                fprintf(fp, "\"");
-
-                xmlRealScopedName(scope, emd->cname, fp);
-
-                fprintf(fp, " const=\"1\" typename=\"int\"/>\n");
+                fprintf(fp, "\" const=\"1\" typename=\"int\"/>\n");
             }
         }
     }
@@ -560,15 +480,13 @@ static void xmlVars(sipSpec *pt, moduleDef *mod, classDef *scope, int indent,
         prScopedPythonName(fp, vd->ecd, vd->pyname->text);
         fprintf(fp, "\"");
 
-        xmlRealName(vd->fqcname, NULL, fp);
-
         if (isConstArg(&vd->type) || scope == NULL)
             fprintf(fp, " const=\"1\"");
 
         if (isStaticVar(vd))
             fprintf(fp, " static=\"1\"");
 
-        xmlType(pt, mod, &vd->type, FALSE, NoKwArgs, fp);
+        xmlType(pt, &vd->type, FALSE, fp);
         fprintf(fp, "/>\n");
     }
 }
@@ -577,57 +495,52 @@ static void xmlVars(sipSpec *pt, moduleDef *mod, classDef *scope, int indent,
 /*
  * Generate the XML for a ctor.
  */
-static void xmlCtor(sipSpec *pt, moduleDef *mod, classDef *scope, ctorDef *ct,
+static int xmlCtor(sipSpec *pt, classDef *scope, ctorDef *ct, int sec,
         int indent, FILE *fp)
 {
-    int a;
+    int a, need_sec;
 
     xmlIndent(indent++, fp);
     fprintf(fp, "<Function name=\"");
     prScopedPythonName(fp, scope, "__init__");
     fprintf(fp, "\"");
 
-    xmlRealScopedName(scope, "__init__", fp);
-
-    if (hasCppSignature(ct->cppsig))
-    {
-        fprintf(fp, " cppsig=\"");
-        xmlCppSignature(fp, ct->cppsig, FALSE);
-        fprintf(fp, "\"");
-    }
-
     /* Handle the trivial case. */
     if (ct->pysig.nrArgs == 0)
     {
         fprintf(fp, "/>\n");
-        return;
+        return FALSE;
     }
 
     fprintf(fp, ">\n");
 
+    need_sec = FALSE;
+
     for (a = 0; a < ct->pysig.nrArgs; ++a)
     {
         argDef *ad = &ct->pysig.args[a];
 
-        if (isInArg(ad))
-            xmlArgument(pt, mod, ad, FALSE, ct->kwargs, FALSE, indent, fp);
+        xmlArgument(pt, ad, dirAttribute(ad), FALSE, sec, indent, fp);
 
-        if (isOutArg(ad))
-            xmlArgument(pt, mod, ad, TRUE, ct->kwargs, FALSE, indent, fp);
+        if (ad->atype == rxcon_type || ad->atype == rxdis_type)
+            need_sec = TRUE;
     }
 
     xmlIndent(--indent, fp);
     fprintf(fp, "</Function>\n");
+
+    return need_sec;
 }
 
 
 /*
  * Generate the XML for a function.
  */
-static void xmlFunction(sipSpec *pt, moduleDef *mod, classDef *scope,
-        memberDef *md, overDef *oloads, int indent, FILE *fp)
+static void xmlFunction(sipSpec *pt, classDef *scope, memberDef *md,
+        overDef *oloads, int indent, FILE *fp)
 {
     overDef *od;
+    const char *default_str = "default=\"1\" ";
 
     for (od = oloads; od != NULL; od = od->next)
     {
@@ -642,40 +555,14 @@ static void xmlFunction(sipSpec *pt, moduleDef *mod, classDef *scope,
 
         if (isSignal(od))
         {
-            int a;
-
-            xmlIndent(indent++, fp);
-            fprintf(fp, "<Signal name=\"");
+            xmlIndent(indent, fp);
+            fprintf(fp, "<Signal %sname=\"", default_str);
             prScopedPythonName(fp, scope, md->pyname->text);
-            fprintf(fp, "\"");
+            fprintf(fp, "\" sig=\"");
+            xmlCppSignature(fp, od);
+            fprintf(fp, "\"/>\n");
 
-            xmlRealScopedName(scope, od->cppname, fp);
-
-            if (hasCppSignature(od->cppsig))
-            {
-                fprintf(fp, " cppsig=\"");
-                xmlCppSignature(fp, od->cppsig, FALSE);
-                fprintf(fp, "\"");
-            }
-
-            /* Handle the trivial case. */
-            if (od->pysig.nrArgs == 0)
-            {
-                fprintf(fp, "/>\n");
-                continue;
-            }
-
-            fprintf(fp, ">\n");
-
-            for (a = 0; a < od->pysig.nrArgs; ++a)
-            {
-                argDef *ad = &od->pysig.args[a];
-
-                xmlArgument(pt, mod, ad, FALSE, od->kwargs, FALSE, indent, fp);
-            }
-
-            xmlIndent(--indent, fp);
-            fprintf(fp, "</Signal>\n");
+            default_str = "";
 
             continue;
         }
@@ -689,7 +576,8 @@ static void xmlFunction(sipSpec *pt, moduleDef *mod, classDef *scope,
             isstat = FALSE;
         }
 
-        xmlOverload(pt, mod, scope, md, od, xtnds, isstat, indent, fp);
+        if (xmlOverload(pt, scope, md, od, xtnds, isstat, FALSE, indent, fp))
+            xmlOverload(pt, scope, md, od, xtnds, isstat, TRUE, indent, fp);
     }
 }
 
@@ -697,41 +585,16 @@ static void xmlFunction(sipSpec *pt, moduleDef *mod, classDef *scope,
 /*
  * Generate the XML for an overload.
  */
-static void xmlOverload(sipSpec *pt, moduleDef *mod, classDef *scope,
-        memberDef *md, overDef *od, classDef *xtnds, int stat, int indent,
-        FILE *fp)
+static int xmlOverload(sipSpec *pt, classDef *scope, memberDef *md,
+        overDef *od, classDef *xtnds, int stat, int sec, int indent, FILE *fp)
 {
-    const char *name, *cppname = od->cppname;
-    int a, no_res;
+    int a, need_sec, no_res;
 
     xmlIndent(indent++, fp);
     fprintf(fp, "<Function name=\"");
-
-    if (isReflected(od))
-    {
-        if ((name = reflectedSlot(md->slot)) != NULL)
-            cppname = name;
-        else
-            name = md->pyname->text;
-    }
-    else
-    {
-        name = md->pyname->text;
-    }
-
-    prScopedPythonName(fp, scope, name);
-
+    prScopedPythonName(fp, scope, md->pyname->text);
     fprintf(fp, "\"");
 
-    xmlRealScopedName(scope, cppname, fp);
-
-    if (hasCppSignature(od->cppsig))
-    {
-        fprintf(fp, " cppsig=\"");
-        xmlCppSignature(fp, od->cppsig, isConst(od));
-        fprintf(fp, "\"");
-    }
-
     if (isAbstract(od))
         fprintf(fp, " abstract=\"1\"");
 
@@ -739,11 +602,10 @@ static void xmlOverload(sipSpec *pt, moduleDef *mod, classDef *scope,
         fprintf(fp, " static=\"1\"");
 
     if (isSlot(od))
-        fprintf(fp, " slot=\"1\"");
-
-    if (isVirtual(od))
     {
-        fprintf(fp, " virtual=\"1\"");
+        fprintf(fp, " slot=\"");
+        xmlCppSignature(fp, od);
+        fprintf(fp, "\"");
     }
 
     if (xtnds != NULL)
@@ -753,146 +615,112 @@ static void xmlOverload(sipSpec *pt, moduleDef *mod, classDef *scope,
         fprintf(fp, "\"");
     }
 
-    /* An empty type hint specifies a void return. */
-    if (od->pysig.result.typehint_out != NULL && od->pysig.result.typehint_out->raw_hint[0] == '\0')
-        no_res = TRUE;
-    else
-        no_res = (od->pysig.result.atype == void_type && od->pysig.result.nrderefs == 0);
+    no_res = (od->pysig.result.atype == void_type && od->pysig.result.nrderefs == 0);
 
     /* Handle the trivial case. */
     if (no_res && od->pysig.nrArgs == 0)
     {
         fprintf(fp, "/>\n");
-        return;
+        return FALSE;
     }
 
     fprintf(fp, ">\n");
 
     if (!no_res)
-        xmlArgument(pt, mod, &od->pysig.result, TRUE, NoKwArgs,
-                isResultTransferredBack(od), indent, fp);
+        xmlArgument(pt, &od->pysig.result, "out", isResultTransferredBack(od),
+                FALSE, indent, fp);
+
+    need_sec = FALSE;
 
     for (a = 0; a < od->pysig.nrArgs; ++a)
     {
         argDef *ad = &od->pysig.args[a];
 
-        /*
-         * Ignore the first argument of non-reflected number slots and the
-         * second argument of reflected number slots.
-         */
-        if (isNumberSlot(md) && od->pysig.nrArgs == 2)
-            if ((a == 0 && !isReflected(od)) || (a == 1 && isReflected(od)))
-                continue;
+        /* Ignore the first argument of number slots. */
+        if (isNumberSlot(md) && a == 0 && od->pysig.nrArgs == 2)
+            continue;
 
-        if (isInArg(ad))
-            xmlArgument(pt, mod, ad, FALSE, od->kwargs, FALSE, indent, fp);
+        xmlArgument(pt, ad, dirAttribute(ad), FALSE, sec, indent, fp);
 
-        if (isOutArg(ad))
-            xmlArgument(pt, mod, ad, TRUE, od->kwargs, FALSE, indent, fp);
+        if (ad->atype == rxcon_type || ad->atype == rxdis_type)
+            need_sec = TRUE;
     }
 
     xmlIndent(--indent, fp);
     fprintf(fp, "</Function>\n");
-}
 
-
-/*
- * Return TRUE if there is a C/C++ signature.
- */
-static int hasCppSignature(signatureDef *sd)
-{
-    int a;
-
-    if (sd == NULL)
-        return FALSE;
-
-    /*
-     * See if there are any arguments that could only have come from
-     * handwritten code.
-     */
-    for (a = 0; a < sd->nrArgs; ++a)
-    {
-        switch (sd->args[a].atype)
-        {
-        case pyobject_type:
-        case pytuple_type:
-        case pylist_type:
-        case pydict_type:
-        case pycallable_type:
-        case pyslice_type:
-        case pytype_type:
-        case pybuffer_type:
-        case capsule_type:
-            return FALSE;
-
-        default:
-            break;
-        }
-    }
-
-    return TRUE;
+    return need_sec;
 }
 
 
 /*
  * Generate the XML for a C++ signature.
  */
-static void xmlCppSignature(FILE *fp, signatureDef *sd, int is_const)
+static void xmlCppSignature(FILE *fp, overDef *od)
 {
-    int a;
-
     prcode(fp, "%M");
-    normaliseArgs(sd);
+    prOverloadDecl(fp, NULL, od, TRUE);
+    prcode(fp, "%M");
+}
 
-    prcode(fp, "(");
 
-    for (a = 0; a < sd->nrArgs; ++a)
+/*
+ * Convert an arguments direction to an XML attribute value.
+ */
+static const char *dirAttribute(argDef *ad)
+{
+    if (isInArg(ad))
     {
-        argDef *ad = &sd->args[a];
+        if (isOutArg(ad))
+            return "inout";
 
-        if (a > 0)
-            prcode(fp, ",");
-
-        generateBaseType(NULL, ad, TRUE, STRIP_GLOBAL, fp);
+        return NULL;
     }
 
-    prcode(fp, ")%s", (is_const ? " const" : ""));
-
-    restoreArgs(sd);
-    prcode(fp, "%M");
+    return "out";
 }
 
 
 /*
  * Generate the XML for an argument.
  */
-static void xmlArgument(sipSpec *pt, moduleDef *mod, argDef *ad, int out,
-        KwArgs kwargs, int res_xfer, int indent, FILE *fp)
+static void xmlArgument(sipSpec *pt, argDef *ad, const char *dir, int res_xfer,
+        int sec, int indent, FILE *fp)
 {
     if (isArraySize(ad))
         return;
 
+    if (sec && (ad->atype == slotcon_type || ad->atype == slotdis_type))
+        return;
+
     xmlIndent(indent, fp);
-    fprintf(fp, "<%s", (out ? "Return" : "Argument"));
-    xmlType(pt, mod, ad, out, kwargs, fp);
+    fprintf(fp, "<Argument");
+    xmlType(pt, ad, sec, fp);
 
-    if (!out)
-    {
-        if (isAllowNone(ad))
-            fprintf(fp, " allownone=\"1\"");
+    if (dir != NULL)
+        fprintf(fp, " dir=\"%s\"", dir);
 
-        if (isDisallowNone(ad))
-            fprintf(fp, " disallownone=\"1\"");
+    if (isAllowNone(ad))
+        fprintf(fp, " allownone=\"1\"");
 
-        if (isTransferred(ad))
-            fprintf(fp, " transfer=\"to\"");
-        else if (isThisTransferred(ad))
-            fprintf(fp, " transfer=\"this\"");
-    }
-
-    if (res_xfer || isTransferredBack(ad))
+    if (isTransferred(ad))
+        fprintf(fp, " transfer=\"to\"");
+    else if (isThisTransferred(ad))
+        fprintf(fp, " transfer=\"this\"");
+    else if (res_xfer || isTransferredBack(ad))
         fprintf(fp, " transfer=\"back\"");
 
+    /*
+     * Handle the default value, but ignore it if it is an output only
+     * argument.
+     */
+    if (ad->defval && (dir == NULL || strcmp(dir, "out") != 0))
+    {
+        prcode(fp, " default=\"%M");
+        exportDefaultValue(ad, FALSE, fp);
+        prcode(fp, "%M\"");
+    }
+
     fprintf(fp, "/>\n");
 }
 
@@ -900,73 +728,69 @@ static void xmlArgument(sipSpec *pt, moduleDef *mod, argDef *ad, int out,
 /*
  * Generate the XML for a type.
  */
-static void xmlType(sipSpec *pt, moduleDef *mod, argDef *ad, int out,
-        KwArgs kwargs, FILE *fp)
+static void xmlType(sipSpec *pt, argDef *ad, int sec, FILE *fp)
 {
-    const char *type_name;
+    const char *type_type = NULL, *type_name;
     classDef *type_scope;
-    typeHintDef *thd;
 
     fprintf(fp, " typename=\"");
 
-    /* Handle the argument name. */
-    if (!out && ad->name != NULL)
+    switch (ad->atype)
     {
-        if (kwargs == AllKwArgs || (kwargs == OptionalKwArgs && ad->defval != NULL))
-            fprintf(fp, "%s: ", ad->name->text);
-    }
+    case class_type:
+        type_type = (isOpaque(ad->u.cd) ? "opaque" : "class");
+        break;
 
-    /* Use any explicit type hint unless the argument is constrained. */
-    thd = (out ? ad->typehint_out : (isConstrained(ad) ? NULL : ad->typehint_in));
+    case enum_type:
+        if (ad->u.ed->pyname != NULL)
+            type_type = "enum";
+        break;
 
-    if (thd != NULL)
-    {
-        pyiTypeHint(pt, thd, mod, out, NULL, FALSE, TRUE, fp);
-    }
-    else
-    {
-        switch (ad->atype)
+    case rxcon_type:
+    case rxdis_type:
+        if (!sec)
+            type_type = "class";
+        break;
+
+    case qobject_type:
+        type_type = "class";
+        break;
+
+    case slotcon_type:
+    case slotdis_type:
         {
-        case class_type:
-            restPyClass(ad->u.cd, fp);
-            break;
+            int a;
 
-        case enum_type:
-            if (ad->u.ed->pyname != NULL)
-                restPyEnum(ad->u.ed, fp);
-            else
-                fprintf(fp, "int");
+            prcode(fp, "SLOT(");
 
-            break;
+            for (a = 0; a < ad->u.sa->nrArgs; ++a)
+            {
+                if (a > 0)
+                    prcode(fp, ", ");
 
-        case qobject_type:
-            restPyClass(pt->qobject_cd, fp);
-            break;
+                prcode(fp, "%M%B%M", &ad->u.sa->args[a]);
+            }
 
-        case mapped_type:
-            /* There would normally be a type hint. */
-            fprintf(fp, "unknown-type");
-            break;
-
-        default:
-            if ((type_name = pyType(pt, ad, &type_scope)) != NULL)
-                prScopedPythonName(fp, type_scope, type_name);
+            prcode(fp, ")");
         }
+
+        break;
+
+    case mapped_type:
+        type_type = "mappedtype";
+        break;
     }
 
-    if (!out && ad->name != NULL && ad->defval != NULL)
-    {
-        fprintf(fp, " = ");
-
-        /*
-         * Try and convert the value to a reST reference.  We don't try very
-         * hard but will get most cases.
-         */
-        if (!restValue(pt, ad->defval, fp))
-            prDefaultValue(ad, FALSE, fp);
-    }
+    if ((type_name = pyType(pt, ad, sec, &type_scope)) != NULL)
+        prScopedPythonName(fp, type_scope, type_name);
 
     fprintf(fp, "\"");
+
+    if (type_type != NULL)
+        fprintf(fp, " typetype=\"%s\"", type_type);
+
+    if (ad->name != NULL)
+        fprintf(fp, " name=\"%s\"", ad->name->text);
 }
 
 
@@ -980,10 +804,42 @@ static void xmlIndent(int indent, FILE *fp)
 }
 
 
+/*
+ * Export the default value of an argument.
+ */
+static void exportDefaultValue(argDef *ad, int in_str, FILE *fp)
+{
+    /* Use any explicitly provided documentation. */
+    if (ad->docval != NULL)
+    {
+        prcode(fp, "%s", ad->docval);
+        return;
+    }
+
+    /* Translate some special cases. */
+    if (ad->defval->next == NULL && ad->defval->vtype == numeric_value)
+    {
+        if (ad->nrderefs > 0 && ad->defval->u.vnum == 0)
+        {
+            prcode(fp, "None");
+            return;
+        }
+
+        if (ad->atype == bool_type || ad->atype == cbool_type)
+        {
+            prcode(fp, ad->defval->u.vnum ? "True" : "False");
+            return;
+        }
+    }
+
+    generateExpression(ad->defval, in_str, fp);
+}
+
+
 /*
  * Get the Python representation of a type.
  */
-static const char *pyType(sipSpec *pt, argDef *ad, classDef **scope)
+static const char *pyType(sipSpec *pt, argDef *ad, int sec, classDef **scope)
 {
     const char *type_name;
 
@@ -1083,10 +939,6 @@ static const char *pyType(sipSpec *pt, argDef *ad, classDef **scope)
 
     switch (ad->atype)
     {
-    case capsule_type:
-        type_name = scopedNameTail(ad->u.cap);
-        break;
-
     case struct_type:
     case void_type:
         type_name = "sip.voidptr";
@@ -1102,27 +954,37 @@ static const char *pyType(sipSpec *pt, argDef *ad, classDef **scope)
             type_name = "int";
         break;
 
+    case signal_type:
+        type_name = "SIGNAL()";
+        break;
+
+    case slot_type:
+        type_name = "SLOT()";
+        break;
+
+    case rxcon_type:
+    case rxdis_type:
+        if (sec)
+            type_name = "callable";
+        else
+            type_name = "QObject";
+
+        break;
+
     case qobject_type:
         type_name = "QObject";
         break;
 
     case ustring_type:
-        /* Correct for Python v3. */
-        type_name = "bytes";
-        break;
-
     case string_type:
     case sstring_type:
     case wstring_type:
     case ascii_string_type:
     case latin1_string_type:
     case utf8_string_type:
-        type_name = isArray(ad) ? "bytes" : "str";
+        type_name = "str";
         break;
 
-    case byte_type:
-    case sbyte_type:
-    case ubyte_type:
     case ushort_type:
     case uint_type:
     case long_type:
@@ -1132,8 +994,6 @@ static const char *pyType(sipSpec *pt, argDef *ad, classDef **scope)
     case short_type:
     case int_type:
     case cint_type:
-    case ssize_type:
-    case size_type:
         type_name = "int";
         break;
 
@@ -1177,14 +1037,15 @@ static const char *pyType(sipSpec *pt, argDef *ad, classDef **scope)
         type_name = "type";
         break;
 
-    case pybuffer_type:
-        type_name = "buffer";
-        break;
-
     case ellipsis_type:
         type_name = "...";
         break;
 
+    case slotcon_type:
+    case anyslot_type:
+        type_name = "SLOT()";
+        break;
+
     default:
         type_name = NULL;
     }
@@ -1193,23 +1054,31 @@ static const char *pyType(sipSpec *pt, argDef *ad, classDef **scope)
 }
 
 
+/*
+ * Generate a scoped Python name.
+ */
+void prScopedPythonName(FILE *fp, classDef *scope, const char *pyname)
+{
+    if (scope != NULL)
+    {
+        prScopedPythonName(fp, scope->ecd, NULL);
+        fprintf(fp, "%s.", scope->pyname->text);
+    }
+
+    if (pyname != NULL)
+        fprintf(fp, "%s", pyname);
+}
+
+
 /*
  * Generate a Python signature.
  */
-static void exportPythonSignature(sipSpec *pt, FILE *fp, signatureDef *sd,
-        int names, int defaults, int in_str, int is_signal)
+int prPythonSignature(sipSpec *pt, FILE *fp, signatureDef *sd, int sec,
+        int names, int defaults, int in_str)
 {
-    int need_comma = FALSE, is_res, nr_out, a;
+    int need_sec = FALSE, need_comma = FALSE, is_res, nr_out, a;
 
-    if (is_signal)
-    {
-        if (sd->nrArgs != 0)
-            fprintf(fp, "[");
-    }
-    else
-    {
-        fprintf(fp, "(");
-    }
+    fprintf(fp, "(");
 
     nr_out = 0;
 
@@ -1223,20 +1092,14 @@ static void exportPythonSignature(sipSpec *pt, FILE *fp, signatureDef *sd,
         if (!isInArg(ad))
             continue;
 
-        need_comma = apiArgument(pt, ad, FALSE, need_comma, names, defaults,
-                in_str, fp);
-    }
+        need_comma = apiArgument(pt, ad, FALSE, need_comma, sec, names,
+                defaults, in_str, fp);
 
-    if (is_signal)
-    {
-        if (sd->nrArgs != 0)
-            fprintf(fp, "]");
-    }
-    else
-    {
-        fprintf(fp, ")");
+        if (ad->atype == rxcon_type || ad->atype == rxdis_type)
+            need_sec = TRUE;
     }
 
+    fprintf(fp, ")");
 
     is_res = !((sd->result.atype == void_type && sd->result.nrderefs == 0) ||
             (sd->result.doctype != NULL && sd->result.doctype[0] == '\0'));
@@ -1249,7 +1112,7 @@ static void exportPythonSignature(sipSpec *pt, FILE *fp, signatureDef *sd,
             fprintf(fp, "(");
 
         if (is_res)
-            need_comma = apiArgument(pt, &sd->result, TRUE, FALSE, FALSE,
+            need_comma = apiArgument(pt, &sd->result, TRUE, FALSE, sec, FALSE,
                     FALSE, in_str, fp);
         else
             need_comma = FALSE;
@@ -1260,201 +1123,13 @@ static void exportPythonSignature(sipSpec *pt, FILE *fp, signatureDef *sd,
 
             if (isOutArg(ad))
                 /* We don't want the name in the result tuple. */
-                need_comma = apiArgument(pt, ad, TRUE, need_comma, FALSE,
+                need_comma = apiArgument(pt, ad, TRUE, need_comma, sec, FALSE,
                         FALSE, in_str, fp);
         }
 
         if ((is_res && nr_out > 0) || nr_out > 1)
             fprintf(fp, ")");
     }
-}
 
-
-/*
- * Generate a fully qualified class name as a reST reference.
- */
-void restPyClass(classDef *cd, FILE *fp)
-{
-    fprintf(fp, ":sip:ref:`~%s.", cd->iff->module->fullname->text);
-    prScopedPythonName(fp, cd->ecd, cd->pyname->text);
-    fprintf(fp, "`");
-}
-
-
-/*
- * Generate a fully qualified enum name as a reST reference.
- */
-void restPyEnum(enumDef *ed, FILE *fp)
-{
-    fprintf(fp, ":sip:ref:`~%s.", ed->module->fullname->text);
-    prScopedPythonName(fp, ed->ecd, ed->pyname->text);
-    fprintf(fp, "`");
-}
-
-
-/*
- * Generate a fully qualified attribute name as a reST reference.
- */
-static void restPyEnumMember(enumMemberDef *emd, FILE *fp)
-{
-    fprintf(fp, ":sip:ref:`~%s.", emd->ed->module->fullname->text);
-    prScopedPythonName(fp, emd->ed->ecd, emd->ed->pyname->text);
-    fprintf(fp, ".%s`", emd->pyname->text);
-}
-
-
-/*
- * Generate a fully qualified attribute name as a reST reference.
- */
-static void restPyAttribute(moduleDef *mod, classDef *scope, nameDef *name,
-        FILE *fp)
-{
-    fprintf(fp, ":sip:ref:`~%s.", mod->fullname->text);
-    prScopedPythonName(fp, scope, name->text);
-    fprintf(fp, "`");
-}
-
-
-/*
- * Generate a reST reference for a scoped name if possible.  Return TRUE if
- * something was generated.
- */
-static int restValue(sipSpec *pt, valueDef *value, FILE *fp)
-{
-    const char *name;
-    scopedNameDef *target, *scope, *snd;
-    varDef *vd;
-    enumDef *ed;
-
-    /* The value must be a scoped name and we don't handle expressions. */
-    if (value->vtype != scoped_value || value->next != NULL)
-        return FALSE;
-
-    target = value->u.vscp;
-
-    /* See if it is an attribute. */
-    for (vd = pt->vars; vd != NULL; vd = vd->next)
-        if (compareScopedNames(vd->fqcname, target) == 0)
-        {
-            restPyAttribute(vd->module, vd->ecd, vd->pyname, fp);
-
-            return TRUE;
-        }
-
-    /* Get the name and scope. */
-    name = scopedNameTail(target);
-
-    scope = NULL;
-
-    for (snd = target; snd->name != name; snd = snd->next)
-        appendScopedName(&scope, text2scopePart(snd->name));
-
-    /* See if it is an enum member. */
-    for (ed = pt->enums; ed != NULL; ed = ed->next)
-    {
-        enumMemberDef *emd;
-
-        /*
-         * Look for the member name first before working out if it is the
-         * correct enum.
-         */
-        for (emd = ed->members; emd != NULL; emd = emd->next)
-        {
-            if (strcmp(emd->cname, name) == 0)
-            {
-                if (isScopedEnum(ed))
-                {
-                    /*
-                     * It's a scoped enum so the fully qualified name of the
-                     * enum must match the scope of the name.
-                     */
-                    if (scope != NULL && compareScopedNames(ed->fqcname, scope) == 0)
-                    {
-                        restPyEnumMember(emd, fp);
-
-                        freeScopedName(scope);
-
-                        return TRUE;
-                    }
-                }
-                else
-                {
-                    /*
-                     * It's a traditional enum so the scope of the enum must
-                     * match the scope of the name.
-                     */
-                    if ((ed->ecd == NULL && scope == NULL) || (ed->ecd != NULL && scope != NULL && compareScopedNames(ed->ecd->iff->fqcname, scope) == 0))
-                    {
-                        if (ed->fqcname == NULL)
-                            restPyAttribute(ed->module, ed->ecd, emd->pyname,
-                                    fp);
-                        else
-                            restPyEnumMember(emd, fp);
-
-                        freeScopedName(scope);
-
-                        return TRUE;
-                    }
-                }
-
-                break;
-            }
-        }
-    }
-
-    freeScopedName(scope);
-
-    return FALSE;
-}
-
-
-/*
- * Return the name of the reflected version of a slot or NULL if it doesn't
- * have one.
- */
-static const char *reflectedSlot(slotType st)
-{
-    switch (st)
-    {
-    case add_slot:
-        return "__radd__";
-
-    case sub_slot:
-        return "__rsub__";
-
-    case mul_slot:
-        return "__rmul__";
-
-    case matmul_slot:
-        return "__rmatmul__";
-
-    case truediv_slot:
-        return "__rtruediv__";
-
-    case floordiv_slot:
-        return "__rfloordiv__";
-
-    case mod_slot:
-        return "__rmod__";
-
-    case lshift_slot:
-        return "__rlshift__";
-
-    case rshift_slot:
-        return "__rrshift__";
-
-    case and_slot:
-        return "__rand__";
-
-    case or_slot:
-        return "__ror__";
-
-    case xor_slot:
-        return "__rxor__";
-
-    default:
-        break;
-    }
-
-    return NULL;
+    return need_sec;
 }
